/*
  main.c 

  (c) J. J. Green 2002
  $Id: main.c,v 1.23 2008/10/17 22:23:45 jjg Exp $
*/

#include <stdio.h>
#include <stdlib.h>
#include <png.h>
#include <errno.h>

#include "config.h"
#include "options.h"
#include "tmask.h"

#define ERROR_OK         0
#define ERROR_USER       1
#define ERROR_READ_OPEN  2
#define ERROR_WRITE_OPEN 3
#define ERROR_MALLOC     4
#define ERROR_BUG        5
#define ERROR_LIBPNG     6

#define JUST_RIGHT 0
#define JUST_LEFT  1

#define DIGITS_UNSET -1

typedef struct opt_t
{
  int   pixels;
  char *input;
  struct
  {
    char *imgpath;
    char *base;
    int digits;
  } name;
  struct
  {
    double left,right;
  } margin;
  int   external;
  int   justify;
  char *tmask;
  int   verbose;
  int   xhtml;
  char *alt;
} opt_t;

static int   process(opt_t);
static char* basename(const char*);
static int   get_options(int,char* const*,opt_t*);

int main(int argc,char* const* argv)
{
  int err;
  opt_t opt;

  err = get_options(argc,argv,&opt);

  if (err == ERROR_OK)
    { 
      if (opt.verbose)
	{
	  printf("This is %s (version %s)\n",
		 OPTIONS_PACKAGE,OPTIONS_VERSION);
	  printf("input file %s\n",opt.input);
	  printf("margins are %g left, %g right\n",
		 opt.margin.left,opt.margin.right);
	}
      err = process(opt);
    }

  if (err)
    {
      char* msg;

      switch (err)
	{
	case ERROR_USER:       msg = "user error"; break;  
	case ERROR_READ_OPEN:  msg = "failed to read file"; break;  
	case ERROR_WRITE_OPEN: msg = "failed to write file"; break;  
	case ERROR_MALLOC:     msg = "out of memory"; break;  
	case ERROR_BUG:        msg = "some kind of bug"; break;  
	case ERROR_LIBPNG:     msg = "error from libpng call"; break;  
	default:               msg = "unknown error - wierd";
	}

      fprintf(stderr,"turmoil! (%s)\n",msg);

      return EXIT_FAILURE;
    }

  if (opt.verbose)
    printf("done.\n");

  return EXIT_SUCCESS;
}

static int get_options(int argc,char* const* argv,opt_t* opt)
{
  struct gengetopt_args_info info;

  options(argc,argv,&info);

  if (info.help_given)    options_print_help();
  if (info.version_given) options_print_version();

  if (info.inputs_num != 1)
    {
      options_print_help();
      return ERROR_USER;
    }
  else opt->input=info.inputs[0];

  opt->pixels  = info.pixels_arg;
  opt->verbose = info.verbose_given;
  opt->xhtml   = info.xhtml_given;

  if (info.left_given && info.right_given)
    {
      fprintf(stderr,"sorry, cannot justify left and right\n");
      return ERROR_USER;
    }

  opt->name.imgpath = 
    (info.imgpath_given ? info.imgpath_arg : NULL);  
      
  opt->justify = 
    (info.right_given ? JUST_RIGHT : JUST_LEFT);

  opt->external = info.external_given;
  
  opt->name.base = 
    (info.basename_given ? info.basename_arg :  basename(opt->input));

  opt->name.digits =
    (info.digits_given ? info.digits_arg : DIGITS_UNSET);

  if (info.margins_given)
    {
      char *left,*right;
      
      if ((left = strdup(info.margins_arg)) == NULL)
	return ERROR_MALLOC;

      if ((right = strchr(left,'/')) != NULL)
	{
	  opt->margin.left  = atof(left);
	  opt->margin.right = atof(right + 1);
	}
      else
	opt->margin.right = opt->margin.left = atof(left);

      free(left);
    }
  else
    {
      opt->margin.left  = 0.0;
      opt->margin.right = 0.0;
    }

  opt->tmask = (info.trim_given ? info.trim_arg : NULL);

  opt->alt = (info.alt_given ? info.alt_arg : NULL);

  return ERROR_OK;
}

/* 
   return a copy of a string up to the last dot,
   (minor memory leak) 
*/

static char* basename(const char* name)
{
  char *base,*last;

  base = strdup(name);
  if ((last = strrchr(base,'.')) != NULL) *last = '\0';

  return base;
}

typedef struct stripdat_t
{
  int n,*cut;
} stripdat_t;

static int get_stripdata(int,int,stripdat_t*);
static int intlog10(int);
static png_bytep* subrows(int,png_bytep*);
static char* create_stystr(opt_t); 
static int create_slice(png_structp,png_infop,png_structp,
			png_infop,png_uint_32,png_bytep*,
			tmask_t,opt_t);

static int process(opt_t opt)
{
  FILE *instream,*htmlstream;
  int  len,bitdepth;
  char *slicefile,*slicename,*htmlfile,*stystr;
  stripdat_t stripdat;
  int i,width,height,channels;
  png_structp ipng,opng; 
  png_infop ipnginfo,opnginfo;
  png_bytep* rows;
  tmask_t tmask;

  /* png input */

  if ((instream = fopen(opt.input,"r")) == NULL)
    {
      fprintf(stderr,"failed to open %s (%s)\n",opt.input,strerror(errno));
      return ERROR_READ_OPEN;
    }

  /* html output */

  if ((htmlfile = malloc(strlen(opt.name.base) + 13)) == NULL)
    return ERROR_MALLOC;

  sprintf(htmlfile,"%s.sliced.html",opt.name.base);

  if ((htmlstream = fopen(htmlfile,"w")) == NULL)
    {
      fprintf(stderr,"failed to open %s\n",htmlfile);
      return ERROR_WRITE_OPEN;
    }

  /* the style string */

  if ((stystr = create_stystr(opt)) == NULL)
    return ERROR_MALLOC;

  /* initalise pnglib & read the ipng */

  ipng = png_create_read_struct(PNG_LIBPNG_VER_STRING,NULL,NULL,NULL);
  if (!ipng)
    {
      fprintf(stderr,"failed to initailise\n");
      return ERROR_LIBPNG;
    }

  ipnginfo = png_create_info_struct(ipng);
  if (!ipnginfo)
    {
      fprintf(stderr,"failed to initailise info\n");
      return ERROR_LIBPNG;
    }

  png_init_io(ipng,instream);
  png_read_png(ipng,ipnginfo,0,NULL);

  /* slice size caculations */

  height = png_get_image_height(ipng,ipnginfo);
  width  = png_get_image_width(ipng,ipnginfo);

  if (get_stripdata(height,opt.pixels,&stripdat) != 0)
    return ERROR_BUG;

  if (opt.verbose)
    printf("image is %i x %i (%i strips)\n",width,height,stripdat.n);

  /* we can now calculate the number of digits (if not specified) */

  if (opt.name.digits == DIGITS_UNSET) opt.name.digits = intlog10(stripdat.n);

  /* 
     convert palettes to rgb (it may be better to make
     a new palette for each output image)
  */

  switch (png_get_color_type(ipng,ipnginfo))
    {
    case PNG_COLOR_TYPE_PALETTE:
    case PNG_COLOR_MASK_PALETTE:
      png_set_palette_to_rgb(ipng);
      if (opt.verbose)
	printf("palette converted to RGB\n");
      break;
      //default:
      /* no palette, nothing to do */
    }

  channels = png_get_channels(ipng,ipnginfo);
  bitdepth = png_get_bit_depth(ipng,ipnginfo);

  if (opt.verbose)
    printf("image has %i channels, depth %i\n",
	   (int)channels,
	   (int)bitdepth);

  if (opt.tmask)
    {
      int tchan;

      tmask = tmask_parse(opt.tmask,bitdepth); 
      tchan = tmask_channels(&tmask);

      if (channels != tchan)
	{
	  fprintf(stderr,
		  "%i channel image and %i channel mask!\n",
		  channels,
		  tchan);

	  return ERROR_USER;
	}
    }
  else
    {
      const char* tsd;

      tsd = tmask_string_default(channels);
      if (opt.verbose) printf("using trim mask %s\n",tsd);

      tmask = tmask_parse(tsd,bitdepth);
      if (!tmask_valid(&tmask)) return ERROR_BUG;
    }

  /* get row pointer array */

  rows = png_get_rows(ipng,ipnginfo);

  /* 
     allocate for the filename used in the html fragment 
     slice filename (slicename) and html and the actual 
     name of the file (slicefile).  slicefile will never
     be longer than slicename, since it is a substring,
     so we use the length of slicename for both.
  */

  len = 
    (opt.name.imgpath ? strlen(opt.name.imgpath) + 1 : 0) + 
    strlen(opt.name.base) + 
    opt.name.digits + 6;

  if ((slicefile = malloc(len)) == NULL || 
      (slicename = malloc(len)) == NULL) return ERROR_MALLOC;

  /* process the slices */

  for (i=0 ; i<stripdat.n ; i++)
    {
      FILE      *outstream;
      int        first,last,nslice,slicewidth;
      png_bytep *slicerows;
 
      first  = stripdat.cut[i];
      last   = stripdat.cut[i+1]-1;
      nslice = last - first + 1;

      /* the name of the file we generate */

      sprintf(slicefile,"%s.%.*i.png",
	      opt.name.base,
	      opt.name.digits,
	      i+1);

      /* the name used in the html fragment */

      if (opt.name.imgpath)
	{
	  int n;

	  /* sprintf return chars printed excluding the terminating null */

	  if ((n=sprintf(slicename,"%s/%s",opt.name.imgpath,slicefile)) != len-1)
	    {
	      fprintf(stderr,"alloacted %i, used %i\n",len,n);
	      return ERROR_BUG;
	    }
	}
      else
	strcpy(slicename,slicefile);

      if ((outstream=fopen(slicefile,"w")) == NULL)
	{
	  fprintf(stderr,"couldnt open %s to write\n",slicefile);
	  return ERROR_WRITE_OPEN;
	}

      /* prepare alt string */

      /* initialise the ouput png structures */
      
      opng = png_create_write_struct(PNG_LIBPNG_VER_STRING,NULL,NULL,NULL);
      if (!opng) return ERROR_LIBPNG;
      
      setjmp(png_jmpbuf(opng));
      png_init_io(opng,outstream);

      opnginfo = png_create_info_struct(opng);
      if (!opnginfo) return ERROR_LIBPNG;

      if (opt.verbose)
	printf("  %s ",slicefile);

      slicerows  = subrows(nslice,rows+first);
      slicewidth = create_slice(ipng,ipnginfo,opng,opnginfo,nslice,slicerows,tmask,opt);

      /* write */

      png_write_png(opng,opnginfo,PNG_TRANSFORM_IDENTITY,NULL);

      /* prepare the alt string */

      size_t alen = 128;
      char alt[alen];

      if (i)
	snprintf(alt,alen,"%i",i+1);
      else
	snprintf(alt,alen,"%s",(opt.alt ? opt.alt : "sliced image"));

      /* write html info */

      fprintf(htmlstream,
	      "<img height=\"%i\" width=\"%i\" src=\"%s\" alt=\"%s\" %s%s\n",
	      nslice,
	      slicewidth,
	      slicename,
	      alt,
	      stystr,
	      (opt.xhtml ? " />" : ">")
	      );

      /* tidy up */

      free(slicerows);

      png_destroy_write_struct(&opng,&opnginfo);
      
      fclose(outstream);
    }

  free(slicefile);
  free(slicename);
  free(stystr);

  png_destroy_read_struct(&ipng,&ipnginfo,NULL);
  fclose(htmlstream);
  fclose(instream);

  return ERROR_OK;
}

/*
  unless the user has specified an external style, we 
  create a fragment of style to be embeded within the 
  attributes if the image. 

  the return value should be freed
*/

/* this is a fascist limit, and should be removed */

#define BUFSIZE 1000

static char* create_stystr(opt_t opt)
{
  char buffer[BUFSIZE];

  if (opt.external)
    buffer[0] = '\0';
  else
    {
      char *juststr;

      juststr = (opt.justify == JUST_RIGHT ? "right" : "left" );

      if (
	  snprintf(
		   buffer,
		   (size_t)BUFSIZE,
		   "style=\"margin-left:%.4gem; margin-right:%.4gem; float:%s; clear:%s\"",
		   opt.margin.left,
		   opt.margin.right,
		   juststr,
		   juststr
		   ) == -1
	  ) return NULL;
    }

  return strdup(buffer);
} 

/*
  integer log base 10 (there are better ways of doing this)
*/

static int intlog10(int n)
{
  int m=0;

  while (n>0)
    {
      n /= 10;
      m++;
    }

  return m;
}

static int get_stripdata(int height,int pixels,stripdat_t *stripdat)
{
  int i,n,*cut;

  n = height/pixels;

  if ((cut=malloc((n+1)*sizeof(int))) == NULL) return ERROR_MALLOC; 
      
  for (i=0 ; i<=n ; i++) cut[i] = i*height/n;

  stripdat->n   = n;
  stripdat->cut = cut;

  return ERROR_OK;
}

/*
  makes a copy of the first n rows (so that we can modify
  them in create_slice()) 
*/

static png_bytep* subrows(int n,png_bytep* rows)
{
  png_bytep* sub;
  int i;

  if ((sub = malloc(n*sizeof(png_bytep))) == NULL)
    return NULL;

  for (i=0 ; i<n ; i++) sub[i] = rows[i];

  return sub;
}

static int get_trim_right(png_bytep*,png_uint_32,png_uint_32,tmask_t);
static int get_trim_left(png_bytep*,png_uint_32,png_uint_32,tmask_t);

static int create_slice(png_structp ipng,
			png_infop ipnginfo,
			png_structp opng,
			png_infop opnginfo,
			png_uint_32 sheight,
			png_bytep* srows,
			tmask_t tmask,
			opt_t opt)
{
  png_uint_32 width,height;
  int bit_depth,color_type,interlace_type,
    compression_type,filter_method,trim;
  
  png_get_IHDR(ipng,
	       ipnginfo,
	       &width,
	       &height,
	       &bit_depth,
	       &color_type,
	       &interlace_type,
	       &compression_type,
	       &filter_method);

  if (opt.justify == JUST_RIGHT)
    {
      int i,nchan;

      trim   = get_trim_left(srows,sheight,width,tmask);
      width -= trim;

      nchan = tmask_channels(&tmask);

      for (i=0 ; i<sheight ; i++) srows[i] += trim*nchan;
    }
  else
    {
      trim = get_trim_right(srows,sheight,width,tmask);
      width -= trim;
    }

  if (opt.verbose)
    printf(" (%i x %i)\n",(int)width,(int)sheight);

  png_set_IHDR(opng,
	       opnginfo,
	       width,
	       sheight,
	       bit_depth,
	       color_type,
	       interlace_type,
	       compression_type,
	       filter_method);

  png_set_rows(opng,opnginfo,srows);

  return width;
}  

static int get_trim_right(png_bytep* rows,
			  png_uint_32 height,
			  png_uint_32 width,
			  tmask_t tmask)

{
  int i,j,min,nchan;

  nchan = tmask_channels(&tmask); 

  min = width;

  for (i=0 ; i<height ; i++)
    {
      png_bytep row;

      row = rows[i];

      for (j=0 ; j<width ; j++)
	{
	  png_bytep pixel;	  
	  
	  pixel = row + nchan*(width - j - 1);

	  if (tmask_match(tmask,pixel) == 0) break;
	}

      if (j<min) min = j;
    }

  return min;
}

static int get_trim_left(png_bytep* rows,
			  png_uint_32 height,
			  png_uint_32 width,
			  tmask_t tmask)

{
  int i,j,min,nchan;

  nchan = tmask_channels(&tmask); 

  min = width;

  for (i=0 ; i<height ; i++)
    {
      png_bytep row;

      row = rows[i];

      for (j=0 ; j<width ; j++)
	{
	  png_bytep pixel;	  
	  
	  pixel = row + nchan*j;

	  if (tmask_match(tmask,pixel) == 0) break;
	}

      if (j<min) min = j;
    }

  return min;
}
